import numpy as np
from numba import njit

def bollinger_bands(bar_data, lookback=20, num_std_dev=2):

    @njit  # Enable Numba JIT.
    def vec_bollinger(values, lookback, num_std_dev):
        n = len(values)
        upper_band = np.array([np.nan for _ in range(n)])
        lower_band = np.array([np.nan for _ in range(n)])
        middle_band = np.array([np.nan for _ in range(n)])

        for i in range(lookback - 1, n):
            window = values[i - lookback + 1:i + 1]
            mean = np.mean(window)
            std = np.std(window)

            middle_band[i] = mean
            upper_band[i] = mean + (std * num_std_dev)
            lower_band[i] = mean - (std * num_std_dev)

        return upper_band, middle_band, lower_band

    return vec_bollinger(bar_data.close, lookback, num_std_dev)


import numpy as np
from numba import njit

def bands_middle(bar_data, lookback=20, num_std_dev=2):

    @njit
    def vec_bollinger(values, lookback, num_std_dev):
        n = len(values)
        middle_band = np.array([np.nan for _ in range(n)])

        for i in range(lookback - 1, n):
            window = values[i - lookback + 1:i + 1]
            mean = np.mean(window)
            middle_band[i] = mean

        return middle_band

    return vec_bollinger(bar_data.close, lookback, num_std_dev)


def bands_upper(bar_data, lookback=20, num_std_dev=2):

    @njit
    def vec_bollinger(values, lookback, num_std_dev):
        n = len(values)
        upper_band = np.array([np.nan for _ in range(n)])

        for i in range(lookback - 1, n):
            window = values[i - lookback + 1:i + 1]
            mean = np.mean(window)
            std = np.std(window)

            upper_band[i] = mean + (std * num_std_dev)

        return upper_band

    return vec_bollinger(bar_data.close, lookback, num_std_dev)


def bands_lower(bar_data, lookback=20, num_std_dev=2):

    @njit
    def vec_bollinger(values, lookback, num_std_dev):
        n = len(values)
        lower_band = np.array([np.nan for _ in range(n)])
        for i in range(lookback - 1, n):
            window = values[i - lookback + 1:i + 1]
            mean = np.mean(window)
            std = np.std(window)

            lower_band[i] = mean - (std * num_std_dev)

        return lower_band

    return vec_bollinger(bar_data.close, lookback, num_std_dev)



def remove_noise_trading(actions=None):
    
    action_open = 0
    action_list = []
    for i, action in enumerate(actions):
        action = int(action)
        action_filter = 0
        if action_open == 0:
            if action != 0:
                action_open = action
                action_filter = action
        elif action_open == 1:
            if action == 1 or action == 0:
                action_filter = 0
            elif action == 2:
                action_filter = action
                action_open = action
        elif action_open == 2:
            if action == 2 or action == 0:
                action_filter = 0
            elif action == 1:
                action_filter = action
                action_open = action
    
        action_list.append(action_filter)
    
    return action_list



def interval_to_minutes(interval):
    if interval.endswith('m'):
        return int(interval[:-1])
    elif interval.endswith('h'):
        return int(interval[:-1]) * 60
    else:
        return None


from ...backtesting_custom.generic_backtesting import GenericBacktesting
from ...utils.pre_processing import remove_noise_trading
from scipy.optimize import minimize
from ta.trend import MACD

class Strategy:
    def __init__(self, df=None):
        pass
    
    def eval(self, df=None, params= [26, 12, 9]):
        
        if len(params) != 3:
            print(f"Print error")
            return False
        
        window_slow, window_fast, window_sign = params
        
        window_slow = int(window_slow)
        window_fast = int(window_fast)
        window_sign = int( window_sign)
        
        print(f"window_slow: {window_slow}, window_fast: {window_fast}, window_sign: {window_sign}")
        # Calcular el indicador MACD
        
        try:
            macd = MACD(df['close'], 
                    window_slow=window_slow, 
                    window_fast=window_fast, 
                    window_sign=window_sign,
                    fillna=True
                ) 
        except:
            print(f"Error en el calculo del MACD")
            return False
        
        macd = MACD(df['close'], 
                    window_slow=window_slow, 
                    window_fast=window_fast, 
                    window_sign=window_sign,
                    fillna=True
                )
        
        df['macd_line'] = macd.macd()
        df['macd_signal'] = macd.macd_signal()
        df.loc[(df['macd_line'] > df['macd_signal']) & (df['macd_line'].shift() <= df['macd_signal'].shift()), 'actions'] = 1  # Señal de compra
        df.loc[(df['macd_line'] < df['macd_signal']) & (df['macd_line'].shift() >= df['macd_signal'].shift()), 'actions'] = 2  # Señal de venta
        df['actions'] = df['actions'].mask(df['actions'].eq(df['actions'].shift()))
        df['actions'] = df['actions'].fillna(0)
        df['actions'] = remove_noise_trading(actions=df['actions'])
        
        
        return df
    
    def objective_function(self, params=[26, 12, 9], df=None):
        
        result_df = self.eval(df=df.copy(), params=params)
        
        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=2000.0,
            commission=0.02,
            plot_result=False,
            path_save_result=".",
            print_stacks=False
        )
        return_data, _ = back_testing.calculate_bk()
        
        return -return_data
    
    def optimize(self, params: list = [(6, 100), (4, 100), (4, 100)], initial_guess: list = [6, 4, 4], df=None, method: str = "Nelder-Mead"):
        threshold_bounds = params
        result = minimize(self.objective_function, initial_guess, bounds=threshold_bounds, args=(df,), method=method)
        best_macd_parameters = result.x
        best_return = -result.fun
        return best_return, best_macd_parameters
    
    def experiment(self, df=None, params=[(6, 100), (4, 100), (4, 100)], initial_guess=[6, 4, 4]):
        
        methods_list = [
            'Nelder-Mead',
            'Powell',
            'CG',
            'L-BFGS-B',
            'COBYLA',
            'trust-constr'
        ]
        results_methods = {}
        results_method_list = []
        result_method_name = []
        
        for method in methods_list:
            print(f"Method: {method}")
            best_return, best_macd_parameters = self.optimize(params=params, initial_guess=initial_guess, df=df, method=method)
            results_method_list.append(best_return)
            result_method_name.append(method)
            results_methods[method] = best_return, best_macd_parameters
        
        # Obtener el máximo valor de results_method_list
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_macd_parameters = results_methods[method]
        
        return method, best_return, best_macd_parameters



from importlib import import_module
import json
import os
from ..backtesting_custom import GenericBacktesting

class ModelSelector: 
    
    def __init__(self, 
                model_name: str= "macd", 
                model_folder: str="strategies",
                type_model: str="basic",
                configuration=None,
                path_model :str ="",
                symbol :str = "",
                df=None,
                interval=None, 
                mlflow=None
            ):
        
        self.mlflow = mlflow
        self.model_name = model_name
        self.model_folder = model_folder
        self.type_model = type_model
        self.configuration = configuration
        self.df = df
        self.path_model = path_model
        self.symbol = symbol
        self.interval = interval
        
        self.strategy_id=f"{self.model_name}_{symbol.lower()}_{interval}"


    def select(self):
        best_return = None
        # BASIC MODEL
        if self.type_model == "basic":
            
            Strategy = self.basic_models()
            
            # EXECUTE OPTIMIZATION IF NOT EXIST INSIDE OF THE FOLDER
            if os.path.isfile(f"{self.path_model}/params.json") == False:      
                strategy = Strategy()
                
                method, best_return, best_sma_threshold = strategy.experiment(
                    df=self.df
                )

                params = {
                    "type": self.type_model,
                    "method": method,
                    "best_return": best_return,
                    "best_sma_threshold": list(best_sma_threshold),
                    "params": list(best_sma_threshold), 
                    "strategy": self.model_name
                }


                self.mlflow.set_tag("strategy", self.model_name)
                self.mlflow.set_tag("method", method)
                self.mlflow.log_param("optimized_params", json.dumps(params))
                self.mlflow.log_metric('best_return', best_return)
                
                df_actions = strategy.eval(df=self.df, params=best_sma_threshold)
                
                back_testing = GenericBacktesting(
                    df=df_actions,
                    skip=True,
                    initial_money=2000.0,
                    commission=0.02,
                    plot_result=False,
                    path_save_result=self.path_model,
                    print_stacks=False
                )

                return_data, stats_json = back_testing.calculate_bk()

                for key, value in stats_json.items():
                    key_s = key.replace(" ",'').replace('[%]','').replace('.','').replace('&','').replace('[$]','').replace('(','').replace(')','').replace('#','')
                    if '_' not in key_s:
                        try:
                            self.mlflow.log_metric(key_s, value)
                        except Exception:
                            pass

                image_folder = self.path_model + "/images"

                if not os.path.exists(image_folder):
                    os.mkdir(image_folder)

            else: 
                # Read params.json
                # MLfow
                params = {}
                # params = json.loads(open(f"{self.path_model}/params.json", "r").read())["best_sma_threshold"]
                
            return Strategy, params, best_return
        
        
        
    def basic_models(self):
        
        libray_trading = f"pyttrading.strategies.{self.model_name}"

        model_module = import_module(libray_trading)
        
        strategy = getattr(model_module, "Strategy")
        
        return strategy
    
    
    def execute(self, df):
        
        Strategy, params, best_return = self.select()
        strategy = Strategy()
        df_action = strategy.eval(df=df, params=params.get('params'))
        
        return best_return, df_action
        
    
    
    
    
     
    
    
    


import json
from backtesting import Backtest, Strategy
import pandas as pd

class DataBacktesting(Strategy):
    """
    Custom strategy class for backtesting.
    """

    def init(self):
        """
        Initialization method called at the beginning of the backtest.
        """
        self.buy_signal = False
        self.sell_signal = False

    def next(self):
        """
        Method called for each data point in the backtest.
        Executes buy or sell based on the 'actions' column value.
        """
        if self.data.actions == 1:
            self.buy()
        elif self.data.actions == 2:
            self.sell()


class GenericBacktesting:
    """
    Generic backtesting class.
    """

    def __init__(self, df, skip=True, initial_money=2000.0, commission=0.02, plot_result=True, path_save_result=".", print_stacks :bool=False):
        """
        Initializes the GenericBacktesting object.

        Parameters:
        - df: DataFrame, input data for backtesting.
        - skip: bool, whether to skip column renaming (default: True).
        - initial_money: float, initial amount of money (default: 2000.0).
        - commission: float, commission rate (default: 0.02).
        - plot_result: bool, whether to plot the backtesting result (default: True).
        - path_save_result: str, path to save the backtesting result (default: ".").
        """
        self.df = df
        self.skip = skip
        self.initial_money = initial_money
        self.commission = commission
        self.plot_result = plot_result
        self.path_save_result = path_save_result
        self.print_stacks = print_stacks

    def rename_columns(self):
        """
        Renames the columns of the input DataFrame.

        The columns 'open', 'height', 'low', 'close', and 'volume' are renamed
        to 'Open', 'High', 'Low', 'Close', and 'Volume', respectively.
        """
        self.df_bk = self.df.rename(columns={
                'open': 'Open',
                'high': 'High',
                'low': 'Low',
                'close': 'Close',
                'volume': 'Volume'
            })

    def calculate_bk(self):
        """
        Calculates the backtesting results and saves them.

        The backtesting results include statistics and optionally plots the result.
        The results are saved in both HTML and JSON formats.
        """
        self.rename_columns()

        self.df_bk_pandas = pd.DataFrame(self.df_bk)
            
        bt = Backtest(self.df_bk_pandas, DataBacktesting,
                    cash=self.initial_money,
                    commission=self.commission,
                    exclusive_orders=True)

        stats = bt.run()

        stats_json = json.loads(stats.to_json()) # Get the value of Return [%]
        
        if self.print_stacks:
            print(stats)
        return stats_json['Return [%]'], stats_json
    
    


def get_backtesting(df=None, initial_money :float = 2000.0):

    back_testing = GenericBacktesting(
                    df=df,
                    skip=True,
                    initial_money=initial_money,
                    commission=0.02,
                    plot_result=False,
                    path_save_result='tmp',
                    print_stacks=False
                )

    return_data, stats_json = back_testing.calculate_bk()
    
    return return_data, stats_json 



from scipy.optimize import minimize
from ..backtesting_custom.generic_backtesting import get_backtesting


def stop_loss_simple(data=None, stop_loss=-0.05, init_capital :float = 2000, column_action_name :str = 'actions', sell_value :int = 2):
    
    price_buy = None
    qty_buy = None
    capital = init_capital

    dff = data.copy()
    for i, action in enumerate(dff['actions']):
        if action == 1:
            # price_buy = dff['close'][i]
            price_buy = dff['close'].iloc[i]
            price_buy = float(price_buy)
            capital = float(capital)

            qty_buy = capital/price_buy

        if qty_buy:
            # current_capital = qty_buy * dff['close'][i]
            current_capital = qty_buy * dff['close'].iloc[i]

            equity = (current_capital-capital)/capital

            if equity < stop_loss:
                dff.loc[dff.index[i], column_action_name] = sell_value

                qty_buy = None
                
                try:
                    filtered_df = dff.iloc[i+1:]
                    idx = filtered_df.index[filtered_df["actions"] == 2][0]
                    dff.at[idx, 'actions'] = 0
                except:
                    pass
        
    return dff


# Optimization Stop Loss

def objective_function(stop_loss, df_original, init_capital=2000, sell_value=2):

    df_stop_loss = stop_loss_simple(data=df_original, stop_loss=stop_loss, init_capital=init_capital, sell_value=sell_value)
    return_data, _ = get_backtesting(df=df_stop_loss)

    print(f"SEARCH THE BEST STOP LOSS SL: {stop_loss} RETURN: {return_data}")
    return -return_data

def optimize_stop_loss_simple(df_original, st_from=0.01, st_to=10):

    #method='Nelder-Mead'
    method='Powell'
    result = minimize(objective_function, x0=0, args=(df_original,), bounds=[(st_from, st_to)], method=method, options={'maxiter': 1000})
    best_stop_loss = result.x[0]
    best_return = -result.fun
    
    return best_stop_loss, best_return



import matplotlib.pyplot as plt


def generate_market_value_plot(df_action_market, df_action_sl_market, best_sl):
    plt.figure(figsize=(8, 4)) 
    plt.plot(df_action_market['market_value'], label='Original')
    plt.plot(df_action_sl_market['market_value'], label=f'SL Opt {best_sl}') 
    plt.legend()
    plt.title('Market Value')
    plt.xlabel('Date') 
    plt.ylabel('Market Value') 
    plt.grid(True)
    plt.savefig('tmp/market_value_plot.png')
    plt.close()  # Close the figure to free memory

    return 'tmp/market_value_plot.png'

def generate_actions_plot(df_action_market):
    
    actions_1 = df_action_market[df_action_market['actions'] == 1]
    actions_2 = df_action_market[df_action_market['actions'] == 2]
    actions_2_sl = df_action_market[df_action_market['actions'] == 2]

    plt.figure(figsize=(10, 6))
    plt.scatter(actions_1.index, actions_1['open'], color='green', label='BUY')
    plt.scatter(actions_2.index, actions_2['open'], color='red', label='SELL')
    plt.scatter(actions_2_sl.index, actions_2_sl['open'], color='orange', label='StopLoss')
    plt.plot(df_action_market.index, df_action_market['close'], label='Close', linestyle='-')
    plt.xlabel('Index')
    plt.ylabel('Values')
    plt.title('Scatter Plot for Actions 1 and 2 with Closing Line')
    plt.legend()
    plt.savefig('tmp/actions_plot.png')
    plt.close()  # Close the figure to free memory

    return 'tmp/actions_plot.png'


import requests
import pandas as pd
import requests
from datetime import datetime, timedelta
from datetime import datetime, timedelta
import pandas as pd
from ..indicators import days_ago
from datetime import datetime
from stockstats import StockDataFrame 


def add_days_ago(data, days):
    for i in range(1, days + 1):
        data[f'd{i}'] = data['close'].shift(i)
    data.dropna(inplace=True)
    return data

class OpensearchCollector:

    def __init__(self,
            url_opensearch: str = "http://localhost:9200",
            url_playground: str = "http://localhost:5005",
            time_amount :str = 10,
            time_unit :str = 'Minute'
            ):

        self.url_opensearch = url_opensearch
        self.url_playground = url_playground
        self.time_amount = time_amount
        self.time_unit = time_unit

    def get_latest_data(self, symbol):
        
        symbol_param = symbol.replace('/', '_')
        url = f"{self.url_opensearch}/{symbol_param.lower()}/_search"

        query = {
            "sort": [
                {"timestamp": {"order": "desc"}}
            ]
        }

        response = requests.get(url, json=query)

        if response.status_code == 200:
            data = response.json()
            hits = data.get("hits", {}).get("hits", [])

            if hits:
                latest_data = hits[0]["_source"]
                df = pd.DataFrame([latest_data])
                date = df['timestamp'][0] 
                date_object = datetime.strptime(date, "%Y-%m-%dT%H:%M:%S%z")


                return date_object
            else:
                print("No se encontraron datos para el rango especificado.")
                return None
        else:
            print(f"Error: {response.status_code}, {response.text}")
            return None


    def data_market_collect(self, symbol, interval, start_date, end_date):
        symbol_param = symbol.replace('/', '_')
        url = f"{self.url_opensearch}/{symbol_param.lower()}/_search"
        # "2023-11-01T09:00:00+00:00",
        query = {
            "query": {
                "range": {
                "timestamp": {
                "gte": start_date.strftime('%Y-%m-%dT%H:%M:%S+00:00'),
                "lt": end_date.strftime('%Y-%m-%dT%H:%M:%S+00:00')
                }
            }
            },
            "aggs": {
                "custom": {
                    "date_histogram": {
                        "field": "timestamp",
                        "interval": interval,
                        "format": "yyyy-MM-dd HH:mm:ss"
                    },
                    "aggs": {
                        "open": {
                            "avg": {
                                "field": "open"
                            }
                        },
                        "high": {
                            "max": {
                                "field": "high"
                            }
                        },
                        "low": {
                            "min": {
                                "field": "low"
                            }
                        },
                        "close": {
                            "avg": {
                                "field": "close"
                            }
                        },
                        "volume": {
                            "sum": {
                                "field": "volume"
                            }
                        },
                        "trade_count": {
                            "sum": {
                                "field": "trade_count"
                            }
                        }
                    }
                }
            }
        }

        response = requests.get(url, json=query)

        if response.status_code == 200:
            data = response.json()
            aggregations = data.get("aggregations", {})
            custom = aggregations.get("custom", {})
            buckets =custom.get("buckets", [])

            data_list = []
            for bucket in buckets:
                data_dict = {
                    "timestamp": bucket.get("key_as_string"),
                    "open": bucket.get("open", {}).get("value"),
                    "high": bucket.get("high", {}).get("value"),
                    "low": bucket.get("low", {}).get("value"),
                    "close": bucket.get("close", {}).get("value"),
                    "volume": bucket.get("volume", {}).get("value"),
                    "trade_count": bucket.get("trade_count", {}).get("value")
                }
                data_list.append(data_dict)

            df = pd.DataFrame(data_list)
            if len(df) > 0:
                df = df.dropna(subset=['open'])
                df = df.rename(columns={'timestamp': 'date'})
                df['symbol'] = symbol
            return df
        else:
            print(f"Error: {response.status_code}, {response.text}")
            return None

    def save_data_opensearch(self, symbols, start_date, end_date, time_amount, time_unit, is_crypto):
        
        url = f"{self.url_playground}/execute"
        if not is_crypto:
            data_type = 'stock'
        else: 
            data_type = 'crypto'
       
            
        params = {
            "symbols": symbols,
            "start_date.year": start_date.year,
            "start_date.month": start_date.month,
            "start_date.day": start_date.day,
            "start_date.hour": start_date.hour,
            "start_date.minute": start_date.minute,
            "end_date.year": end_date.year,
            "end_date.month": end_date.month,
            "end_date.day": end_date.day,
            "end_date.hour": end_date.hour,
            "end_date.minute": end_date.minute,
            "time_amount": time_amount,
            "time_unit": time_unit,
            "data_type": data_type
        }

        
        headers = {'accept': 'application/json'}
        response = requests.get(url, params=params, headers=headers)
        if response.status_code == 200:
            data = response.json()
            return data
        else:
            print(f"Error: {response.status_code}, {response.text}")
            return None

    def market_data_days_ago(self, delay_minutes: int = 80, search_days_ago :int = 300, symbol :str = 'AMZN', data_type='stock', interval='1h', is_crypto=False):

        now_datetime = datetime.now()
        delta = timedelta(minutes=delay_minutes)
        delta_month = timedelta(days=search_days_ago)
        end_date = now_datetime - delta
        start_date = now_datetime - delta_month
        start_date_request = start_date

        latest_date_ops = self.get_latest_data(symbol=symbol)
        if isinstance(latest_date_ops, datetime):
            start_date = latest_date_ops


        self.save_data_opensearch(
            symbols=symbol, 
            start_date=start_date,
            end_date=end_date, 
            time_amount=self.time_amount, 
            time_unit=self.time_unit, 
            is_crypto=is_crypto
        )


        data = self.data_market_collect(
            symbol=symbol, 
            interval=interval,
            start_date=start_date_request,  
            end_date=end_date
            )



        return data
    

    def market_data(self, symbols: str = 'AMZN', start_date: str = '6/1/2021', end_date: str = '12/1/2021', interval: str = '1h', is_crypto=False):


        if ':' not in start_date:
            start_date = datetime.strptime(start_date, '%m/%d/%Y') if isinstance(start_date, str) else start_date
        else:
            start_date = datetime.strptime(start_date, '%m/%d/%Y %H:%M') if isinstance(start_date, str) else start_date

        if ':' not in end_date:
            end_date = datetime.strptime(end_date, '%m/%d/%Y') if isinstance(end_date, str) else end_date
        else:
            end_date = datetime.strptime(end_date, '%m/%d/%Y %H:%M') if isinstance(end_date, str) else end_date

        if start_date.tzinfo is not None:
            start_date = start_date.replace(tzinfo=None)
        if end_date.tzinfo is not None:
            end_date = end_date.replace(tzinfo=None)
       
        start_date_required = start_date
        data_list = []

        if not isinstance(symbols,list):
            symbols = [symbols]
        for symbol in symbols:
    
            latest_date_ops = self.get_latest_data(symbol=symbol)
            if isinstance(latest_date_ops, datetime):
             
                if latest_date_ops.tzinfo is not None:
                    latest_date_ops = latest_date_ops.replace(tzinfo=None)
                start_date = max(start_date, latest_date_ops)
            
            self.save_data_opensearch(
                symbols=symbol, 
                start_date=start_date,
                end_date=end_date, 
                time_amount=self.time_amount, 
                time_unit=self.time_unit, 
                is_crypto=is_crypto
            )

            # Obtener datos del mercado
            data = self.data_market_collect(
                symbol=symbol, 
                interval=interval,
                start_date=start_date_required,  
                end_date=end_date
            )


            if hasattr(self, 'indicators') and self.indicators:
                data = StockDataFrame.retype(data)
                for indicator in self.indicators:
                    data[indicator]

            data_list.append(data)  # Agregar cada DataFrame a la lista

        data_pd = pd.concat(data_list, ignore_index=True)
        data_pd['date'] = pd.to_datetime(data_pd['date'])
        return  data_pd


    def market_data_old(self, symbols :str = 'AMZN', start_date='6/1/2021', end_date='12/1/2021', interval='1h', points_ago :int =4):


        data_list = []  # Lista para almacenar los DataFrames individuales

        self.points_ago = 5
        self.indicators = ['boll_ub', 'boll_lb', 'boll']

        for symbol in symbols:
            self.save_data_opensearch(
                symbols=symbol, 
                start_date=start_date,
                end_date=end_date, 
                time_amount=self.time_amount, 
                time_unit=self.time_unit
            )

            # Resto del código para obtener datos
            data = self.data_market_collect(
                symbol=symbol, 
                interval=interval,
                start_date=start_date,  
                end_date=end_date
            )


            if self.points_ago:
                data = add_days_ago(data=data, days=self.points_ago)
            if self.indicators:
                data = StockDataFrame.retype(data)
        
            data_list.append(data)  # Agregar cada DataFrame a la lista
            
            data = days_ago(data=data, days=5)
            data_list.append(data)  # 


            for indicator in self.indicators:
                data[indicator]

        if data_list:
            concatenated_data = pd.concat(data_list, ignore_index=True)
            concatenated_data['date'] = pd.to_datetime(concatenated_data['date'])

            return concatenated_data
        else:
            return None




from ...backtesting_custom.generic_backtesting import GenericBacktesting
from ...utils.pre_processing import remove_noise_trading
from scipy.optimize import minimize

# this strategy is breakout

class Strategy: 

    def __init__(self, df=None):
        
        pass         
    def eval(self,df=None, params= None):
        
        breakout_threshold = int(params)
        
        print(breakout_threshold)
        # Calcular el rango de ruptura (breakout range)
        df['breakout_range'] = df['height'] - df['low']

        df['actions'] = 0  # Inicializar todas las acciones como "mantener"
        df.loc[df['close'] > df['open'] + breakout_threshold * df['breakout_range'], 'actions'] = 1  # Señal de compra
        df.loc[df['close'] < df['open'] - breakout_threshold * df['breakout_range'], 'actions'] = 2  # Señal de venta

        # Eliminar señales consecutivas repetidas
        df['actions'] = df['actions'].mask(df['actions'].eq(df['actions'].shift()))
        # Reemplazar NaN por 0
        df['actions'] = df['actions'].fillna(0)

        df['actions'] = remove_noise_trading(actions=df['actions'])

        return df
    

    def objective_function(self,params, df=None):

        self.df = df
        result_df = self.eval(df=df, params=params)

        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=2000.0,
            commission=0.02,
            plot_result=False,
            path_save_result="."
        )

        return_data, _ = back_testing.calculate_bk()

        return -return_data
    
    def optimize(self, parms: list =  [(0.1, 1.3)], initial_guess: list = [0.02], df=None, method: str = 'Nelder-Mead'):

        threshold_bounds = parms

        # Realizar la optimización
        result = minimize(self.objective_function, initial_guess, bounds=threshold_bounds, args=(df), method=method)
        # Obtener el mejor valor de breakout_threshold y el mejor retorno
        best_breakout_threshold = result.x[0]
        best_return = -result.fun
        
        return best_return, best_breakout_threshold  
    

    def experiment(self, df=None, parms=[(0.1, 1.3)],initial_guess=[0.02]):
        
        methods_list = [
            'Nelder-Mead',
            'Powell',
            'CG',
            'BFGS',
            'L-BFGS-B',
            # 'TNC',
            'COBYLA',
            'SLSQP',
            'trust-constr',
            # 'dogleg',
            # 'trust-ncg',
            # 'trust-exact',
            # 'trust-krylov'
        ]
        
        results_methods = {}
        results_method_list = []
        result_method_name = []
        
        for method in methods_list:
            print(f"Method: {method}")
    
            best_return, best_sma_threshold = self.optimize(
                parms=parms, 
                initial_guess=initial_guess,
                df=df, 
                method=method
            )
    
            results_method_list.append(best_return)
            result_method_name.append(method)
            results_methods[method] = best_return, best_sma_threshold
            

        # get the max value of results_method_list
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_sma_threshold = results_methods[method]
        
        return method, best_return, best_sma_threshold


from ...backtesting_custom.generic_backtesting import GenericBacktesting
from ...utils.pre_processing import remove_noise_trading
from scipy.optimize import minimize
from ta.trend import SMAIndicator

class Strategy: 
    def __init__(self, df=None):
        pass         

    def eval(self, df=None, params=[10, 30]):
        
        if len(params) != 2:
            print(f"Print error")
            return False
        
        fast_ma_period, slow_ma_period = params
        
        print(f"fast_ma_period: {fast_ma_period} slow_ma_period: {slow_ma_period}")
        # Calcular el rango de ruptura (breakout range)
        df['fast_ma'] = SMAIndicator(df['close'], int(fast_ma_period), True).sma_indicator()
        df['slow_ma']  = SMAIndicator(df['close'], int(slow_ma_period), True).sma_indicator()
        df['actions'] = 0  # Inicializar todas las acciones como "mantener"
        df.loc[df['fast_ma'] > df['slow_ma'], 'actions'] = 1  # Señal de compra
        df.loc[df['fast_ma'] < df['slow_ma'], 'actions'] = 2  # Señal de venta
        # Eliminar señales consecutivas repetidas
        df['actions'] = df['actions'].mask(df['actions'].eq(df['actions'].shift()))
        # Reemplazar NaN por 0
        df['actions'] = df['actions'].fillna(0)
        
        df['actions'] = remove_noise_trading(actions=df['actions'])
        
        
        return df
    
    def objective_function(self, params=[10,30], df=None):
        # Apply the strategy with the current parameters
        
        result_df = self.eval(df=df.copy(), params=params)
        
        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=2000.0,
            commission=0.02,
            plot_result=False,
            path_save_result=".",
            print_stacks=False
        )
        return_data, _ = back_testing.calculate_bk()
        # Minimize the negative return to maximize the return
        return -return_data
    
    def optimize(self, parms: list = [(5, 30), (10, 100)], initial_guess: list = [10, 20], df=None, method: str = "Nelder-Mead"):
        
        threshold_bounds = parms
        
        result = minimize(self.objective_function, initial_guess, bounds=threshold_bounds, args=(df), method=method)
        
        best_breakout_threshold = result
        
        best_return = -result.fun

        return best_return, best_breakout_threshold
    



    def experiment(self, df=None, parms=[(5, 100), (40, 200)],initial_guess=[5, 40]):
        
        methods_list = [
            'Nelder-Mead',
            'Powell',
            'CG',
            # 'BFGS',
            'L-BFGS-B',
            # 'TNC',
            'COBYLA',
            # 'SLSQP',
            'trust-constr',
            # 'dogleg',
            # 'trust-ncg',
            # 'trust-exact',
            # 'trust-krylov'
        ]
        
        results_methods = {}
        results_method_list = []
        result_method_name = []
        
        for method in methods_list:
            print(f"Method: {method}")
    
            best_return, best_sma_threshold = self.optimize(
                parms=parms, 
                initial_guess=initial_guess,
                df=df, 
                method=method
            )
    
            results_method_list.append(best_return)
            result_method_name.append(method)
    
            results_methods[method] = best_return, best_sma_threshold.x
            

        # get the max value of results_method_list
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_sma_threshold = results_methods[method]
        
        return method, best_return, best_sma_threshold


from ...backtesting_custom.generic_backtesting import GenericBacktesting
from ...utils.pre_processing import remove_noise_trading

from scipy.optimize import minimize
from ta.momentum import RSIIndicator

class Strategy:
    
    def __init__(self, df=None):
        pass
    
    def eval(self, df=None, params=[5, 100]):
        
        if len(params) != 2:
            print(f"Print error")
            return False
        
        window_length, buy_threshold = params
        
        # Calculate the RSI indicator
        rsi_indicator = RSIIndicator(df['close'], window=window_length)
        self.df = df
        df['rsi'] = rsi_indicator.rsi()
        df['actions'] = 0  # Initialize the actions column with 0
        df.loc[df['rsi'] < buy_threshold, 'actions'] = 1  # Buy signal
        df.loc[df['rsi'] >= buy_threshold, 'actions'] = 2 # Sell Signal
        df['actions'] = remove_noise_trading(actions=df['actions'])
        
        
        return df
    
    def objective_function(self, params=[14, 30], df=None):
        result_df = self.eval(df=df.copy(), params=params)
        
        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=2000.0,
            commission=0.02,
            plot_result=False,
            path_save_result=".",
            print_stacks=False
        )
        return_data, _ = back_testing.calculate_bk()
        
        return -return_data
    
    def optimize(self, params: list = [(5, 100), (5, 100)], initial_guess: list = [5, 5], df=None, method: str = "Nelder-Mead"):
        threshold_bounds = params
        result = minimize(self.objective_function, initial_guess, bounds=threshold_bounds, args=(df,), method=method)
        best_parameters = result.x
        best_return = -result.fun
        return best_return, best_parameters
    
    def experiment(self, df=None, params=[(5, 100), (5, 100)], initial_guess=[5, 5]):
        methods_list = [
            'Nelder-Mead',
            'Powell',
            'CG',
            'L-BFGS-B',
            'COBYLA',
            'trust-constr'
        ]
        results_methods = {}
        results_method_list = []
        result_method_name = []
        
        for method in methods_list:
            print(f"Method: {method}")
            best_return, best_parameters = self.optimize(params=params, initial_guess=initial_guess, df=df, method=method)
            results_method_list.append(best_return)
            result_method_name.append(method)
            results_methods[method] = best_return, best_parameters
        
        # Get the maximum value from results_method_list
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_parameters = results_methods[method]
        
        return method, best_return, best_parameters



import pandas as pd
import numpy as np

class Strategy:

    def __init__(self, df=None, params=None):
        self.df = df

    def eval(self, df=None, params=None):
        df = df.copy()
        df['actions'] = np.random.choice([0,1,2], size=len(df), p=[0.7, 0.1, 0.2])
        return df

    def objective_function(self, df=None, params=None):
        result_df = self.eval(df=df.copy())
        return result_df['actions'].sum()

    # def optimize(self, df=None, parms=None):
    def optimize(self, parms: list = [(5, 30), (10, 100)], initial_guess: list = [10, 20], df=None, method: str = "Nelder-Mead"):

        best_return = self.objective_function(df=df)
        return best_return

    def experiment(self, df=None, parms=[(5, 100), (40, 200)],initial_guess=[5, 40]):
        best_return = self.optimize(df=df)
        return best_return



from ...backtesting_custom.generic_backtesting import GenericBacktesting
from ...utils.pre_processing import remove_noise_trading
from scipy.optimize import minimize
from ta.trend import EMAIndicator


class Strategy:
    
    def __init__(self, df=None):
        self.df = df

    def eval(self, df=None, params=[20, 2]):
        # window_length, stddev_threshold = map(int, params)
        
        if len(params) != 2:
            return False
        window_length, stddev_threshold = params
        

        # Calculate the EMA indicator
        ema_indicator = EMAIndicator(close=df['close'], window=window_length)
        df['ema'] = ema_indicator.ema_indicator()

        # Initialize the actions column with 0
        df['actions'] = 0

        # Generate buy signals based on EMA indicator
        df.loc[df['close'] > df['ema'], 'actions'] = 1  # Buy signal
        df.loc[df['close'] < df['ema'], 'actions'] = 2  # Sell signal

        df['actions'] = remove_noise_trading(actions=df['actions'])

        return df

    def objective_function(self, params=[20, 2], df=None):
        result_df = self.eval(df=df.copy(), params=params)

        back_testing = GenericBacktesting(
            df=result_df,
            skip=True,
            initial_money=2000.0,
            commission=0.02,
            plot_result=False,
            path_save_result=".",
            print_stacks=False
        )
        return_data, _ = back_testing.calculate_bk()

        return -return_data

    def optimize(self, params=[(2, 50), (1, 10)], initial_guess=[2, 2], df=None, method="Nelder-Mead"):
        result = minimize(self.objective_function, initial_guess, bounds=params, args=(df,), method=method)
        best_parameters = result.x
        best_return = -result.fun
        return best_return, best_parameters

    def experiment(self, df=None, params=[(2, 50), (1, 10)], initial_guess=[2, 2]):
        methods_list = [
            'Nelder-Mead',
            'Powell',
            'CG',
            # 'BFGS',
            'L-BFGS-B',
            'TNC',
            # 'COBYLA',
            'SLSQP',
            'trust-constr',
            # 'dogleg',
            # 'trust-ncg',
            # 'trust-exact',
            # 'trust-krylov'
        ]
        results_methods = {}
        results_method_list = []
        result_method_name = []

        for method in methods_list:
            try:
                print(f"Method: {method}")
                best_return, best_parameters = self.optimize(params=params, initial_guess=initial_guess, df=df, method=method)
                results_method_list.append(best_return)
                result_method_name.append(method)
                results_methods[method] = best_return, best_parameters
            except Exception as e:
                print(f"Error in method {method}: {e}")

        # Get the maximum value from results_method_list
        optimize = results_method_list.index(max(results_method_list))
        method = result_method_name[optimize]
        best_return, best_parameters = results_methods[method]

        return method, best_return, best_parameters



from setuptools import find_packages, setup

setup(
    name='pyttrading',
    packages=find_packages(),
    version='0.6.0',
    description='Trading Library',
    author='Cecilio Cannavacciuolo Diaz',
    install_requires=[],
    setup_requires=[
        'pytest-runner',
        'stock-dataframe==0.1.0',
        'mlflow==2.9.2',
        'backtesting==0.3.3'
    ],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
    python_requires='>=3.6'
)

# pip uninstall -y pyttrading && python setup.py sdist && twine upload dist/* 



import pandas as pd
from pyttrading.strategies.selector import ModelSelector


def test_collect_market_data():

    data = pd.read_csv('data.csv')

    strategy_config  = { 
        "is_crypto": True,
        "symbol": 'AVAX/USD',
        "interval": '2m',
        "start_date": '11/1/2023',
        "end_date": '12/16/2023',
        "indicators": [
            "boll_ub",
            "boll_lb",
            "close_10_sma",
            "close_12_ema",
            "close_16_ema"
        ],
    }

    strategy_config = ModelSelector(
                    model_name='dummy',
                    path_model="tmp",
                    type_model="basic",
                    configuration=strategy_config, 
                    df=data,
                    symbol=strategy_config.get("symbol"),
                    interval=strategy_config.get("interval"),
                    mlflow=None
                )
    
    Strategy = strategy_config.basic_models()
    strategy = Strategy()
    df_actions = strategy.eval(df=data, params=None)


    assert len(df_actions[['actions']] ) > 0


from .myfunctions import haversine
from .ema import Strategy as ema
from .rsi import Strategy as rsi
from .macd import Strategy as macd
from .sma import Strategy as sma
from .dummy import Strategy as dummy